import numpy as np
import xarray as xr


def time_deconding(ds: xr.Dataset, time_name: str = "Time") -> xr.Dataset:

    ds["Time"] = ds["Time"] + ds["Timems"] * 1e-3
    ds["Time"].attrs["units"] = "seconds since 20010101 00:00:00"
    ds = xr.decode_cf(ds)

    return ds


def selecting_variables(dataset):

    velocities_ds = xr.Dataset()
    for chirp_name in ["C1", "C2", "C3"]:

        tmp_data = dataset[f"{chirp_name}MeanVel"].rename(
            {f"{chirp_name}Range": "range"}
        )
        tmp_data.name = "MeanVel"
        velocities_ds = xr.merge([velocities_ds, tmp_data])

        tmp_data = dataset[f"{chirp_name}ZDR"].rename(
            {f"{chirp_name}Range": "range"}
        )
        tmp_data.name = "ZDR"
        velocities_ds = xr.merge([velocities_ds, tmp_data])

    # tmp_rain_data = dataset["Rain"]
    # tmp_rain_data.name = 'rain_rate'

    velocities_ds = velocities_ds.where(velocities_ds != -999)
    velocities_ds = xr.merge(
        [velocities_ds, dataset.Azm, dataset.Elv]
    )  # , tmp_rain_data])
    velocities_ds = velocities_ds.rename(
        {"Time": "time", "Azm": "azimuth", "Elv": "elevation"}
    )
    velocities_ds["range_layers"] = velocities_ds.range
    velocities_ds = velocities_ds.assign_coords(
        {"range": np.arange(len(velocities_ds.range))}
    )

    return velocities_ds


def azimuth_offset(ds, variable="MeanVel"):
    # those offset values where retrieved using
    # the windcube lidar as reference

    result = ds.azimuth[-1] - ds.azimuth[0]

    # increasing azm
    if result > 0:

        # ds[variable].attrs['azmOff_c3'] = '4.88'
        # ds[variable].attrs['azmOff_c2'] = '4.23'
        # ds[variable].attrs['azmOff_c1'] = '2.08'
        ds["azm_seq"] = 1

    # decreasing azm
    if result < 0:

        # ds[variable].attrs['azmOff_c3'] = '0'
        # ds[variable].attrs['azmOff_c2'] = '0'
        # ds[variable].attrs['azmOff_c1'] = '0'
        ds["azm_seq"] = -1

    return ds


def height_estimation(ds):

    # height estimation
    correc_fact = np.sin(np.deg2rad(ds["elevation"].values[0]))
    ds["range_layers"] = ds.range_layers * correc_fact

    return ds


def fill_nan_values(ds):

    """

    It does more than fill the NaNs by interpolation.
    It replaces the time coordinate by the azimuths.
    It also drops some time or azimuth duplicates.

    MDV[range, time] --> MDV[range, azimuth]

    """

    tmp_ds = ds.drop_duplicates(dim="time")
    tmp_ds = tmp_ds.interpolate_na(dim="time")

    return tmp_ds


def update_structure(ds):

    """

    It does more than fill the NaNs by interpolation.
    It replaces the time coordinate by the azimuths.
    It also drops some time or azimuth duplicates.

    MDV[range, time] --> MDV[range, azimuth]

    """

    # tmp_ds = ds.MeanVel.drop_duplicates(dim='time')
    # tmp_ds = tmp_ds.interpolate_na(dim='time')

    tmp_ds = (
        ds.swap_dims({"time": "azimuth"})
        .reset_coords()
        .drop_duplicates(dim="azimuth")
    )
    tmp_ds = tmp_ds.assign_coords({"elevation": ds.elevation.values[0]})

    tmp_ds = tmp_ds.assign_coords({"mean_time": ds.time.mean()})

    ## in the future, I have to use an external function for it
    start_scan = (
        ds.time[0].assign_coords({"mean_time": tmp_ds.mean_time}).drop("time")
    )
    start_scan.name = "start_scan"
    end_scan = (
        ds.time[-1].assign_coords({"mean_time": tmp_ds.mean_time}).drop("time")
    )
    end_scan.name = "end_scan"

    max_zdr = ds.ZDR.max("time").assign_coords({"mean_time": tmp_ds.mean_time})
    max_zdr.name = "zdr_max"

    nan_percent = ds.nan_percentual.assign_coords(
        {"mean_time": tmp_ds.mean_time}
    )

    tmp_ds = xr.merge([tmp_ds, start_scan, end_scan, max_zdr, nan_percent])

    return tmp_ds


def get_chirp_information(ds):

    chirp_start = [ds[f"C{c+1}Range"].values.min() for c in ds["Chirp"].values]
    chirp_end = [ds[f"C{c+1}Range"].values.max() for c in ds["Chirp"].values]

    chirp_start = xr.DataArray(
        chirp_start,
        coords={"chirp": ds["Chirp"].values + 1},
        name="chirp_start",
    )
    chirp_end = xr.DataArray(
        chirp_end, coords={"chirp": ds["Chirp"].values + 1}, name="chirp_end"
    )
    chirp_dir_bias = xr.DataArray(
        np.zeros_like(ds["Chirp"]),
        coords={"chirp": ds["Chirp"].values + 1},
        name="chirp_azimuth_bias",
    )

    return xr.merge([chirp_start, chirp_end, chirp_dir_bias])


def rpg_slanted_radial_velocity_4_fft(ds):

    chirp_info = get_chirp_information(ds)

    # pre-processing
    ds = time_deconding(ds)
    ds = selecting_variables(ds)
    ds = azimuth_offset(ds)
    ds = height_estimation(ds)

    ds = update_range(ds)
    ds = count_nan_values(ds)
    ds = fill_nan_values(ds)

    ds = update_structure(ds)
    ds = azimuth_regular_grid_interp(ds)
    ds = nan_leftover_to_mean(ds)

    ds = ds.merge(chirp_info)

    return ds


def update_range(ds):

    """

    It replaces the original range numbers stored in the
    range coordinate with the values stored in variable
    range_layers. range_layers stores the distance from
    the antenna to the centre of each range gate

    """

    ds["range"] = ds.range_layers

    return ds


def count_nan_values(ds):

    """

    It counts the number of NaNs present in the MDV per
    range gate and returns the percentual of NaN.

    """

    total_nan = ds.time.shape - (ds["MeanVel"] / ds["MeanVel"]).sum(dim="time")
    percent_nan = 100 * total_nan / ds.time.shape
    percent_nan.name = "nan_percentual"

    ds = ds.merge(percent_nan)

    return ds


def azimuth_regular_grid_interp(ds):

    """

    It interpolates MDV[range, azimuth] from a
    irregular azimuth grid to a regular azimuth
    grid. The resolution of the new azimuth coordinate
    is 5 degrees, which is close to the original
    resolution.

    This step is required in order to use the FFT for
    retrieving wind speed and direction.

    """

    azimuths = np.arange(0, 360, 5)  # 5
    ds = ds.interp(azimuth=azimuths)

    return ds


def nan_leftover_to_mean(ds):

    """

    For each range, It replaces all remaining NaNs from
    the interpolated MDV by the mean values.

    This step is also required in order to use the FFT for
    retrieving wind speed and direction.

    """

    mean_rad_vel = ds.MeanVel.mean(dim="azimuth")

    for r in range(ds.range.shape[0]):

        nan_index = ~np.isfinite(ds.MeanVel[:, r])
        ds.MeanVel[:, r][nan_index] = mean_rad_vel[r]

    return ds



import numpy as np
import pandas as pd
import xarray as xr

from lidarwind import preprocessing
from lidarwind.wind_retrieval import fft_wind_retrieval


def sintetic_data(step=90, elevation=75) -> xr.Dataset:
    """Sintect data

    Function to create synthetic data for testing the package

    Parameters
    ----------
    step : int
        separation in degree between the data points

    elevation : int
        elevation of the observation

    Returns
    -------
        A synthetic dataset

    """

    azimuths = np.arange(0, 360, step)

    wind = np.cos(np.deg2rad(azimuths + 0)) * 20 * np.cos(np.deg2rad(75))

    elv = np.append(np.ones_like(azimuths) * elevation, [90, 75])

    time = pd.to_datetime("11/11/2017") + pd.to_timedelta(
        np.arange(len(elv)), unit="seconds"
    )

    data_elv = xr.DataArray(elv, dims=("time"), coords={"time": time})

    azm = np.append(azimuths, [0, 0])
    data_azm = xr.DataArray(azm, dims=("time"), coords={"time": time})

    data = xr.DataArray(
        (np.ones((2, 7)) * np.append(wind, [0, wind[0]])).T,
        dims=("time", "gate_index"),
        coords={"time": time, "gate_index": [1, 2]},
    )

    ranges = xr.DataArray(
        np.ones((2, 7)).T
        * np.array([100, 150])
        / np.sin(np.deg2rad(elevation)),
        dims=("time", "gate_index"),
        coords={"time": time, "gate_index": [1, 2]},
    )

    data_status = xr.DataArray(
        np.ones_like(data.values),
        dims=("time", "gate_index"),
        coords={"time": time, "gate_index": [1, 2]},
    )

    test_ds = xr.Dataset(
        {
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr": data,
            "range": ranges,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data_status,
            "relative_beta": data,
        }
    )

    test_ds = test_ds.set_coords(
        {
            "elevation": test_ds.elevation,
            "azimuth": test_ds.azimuth,
            "range": ranges,
        }
    )

    test_ds["range"].values[test_ds.elevation == 90] = np.array([100, 150])

    return test_ds


def get_radial_velocities_4_test():

    ds = sintetic_data(step=72, elevation=75)
    slanted_elevation = np.unique(ds.elevation)[np.unique(ds.elevation) != 90]
    ds_slanted = ds.where(ds.elevation == slanted_elevation, drop=True)
    radial_velocities = preprocessing.wc_slanted_radial_velocity_4_fft(
        ds_slanted
    )

    return radial_velocities


def test_fisrt_harmonic_amplitude_size():

    tmp_amp = fft_wind_retrieval.first_harmonic_amplitude(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert tmp_amp.size > 0


def test_first_harmonic_amplitude_magnitude():

    tmp_amp = fft_wind_retrieval.first_harmonic_amplitude(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert np.round(np.absolute(tmp_amp.values[1, 1]), 2) == 12.94


def test_first_harmonic_amplitude_direction():

    tmp_amp = fft_wind_retrieval.first_harmonic_amplitude(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert fft_wind_retrieval.wind_direction(tmp_amp).values[1, 1] == 180


def test_first_harmonic_amplitude_wind_speed():

    tmp_amp = fft_wind_retrieval.first_harmonic_amplitude(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert np.isclose(fft_wind_retrieval.wind_speed(tmp_amp).values[1, 1], 20)


def test_first_harmonic_amplitude_meridional_wind():

    tmp_amp = fft_wind_retrieval.first_harmonic_amplitude(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert np.isclose(
        fft_wind_retrieval.meridional_wind(tmp_amp).values[1, 1], 20
    )


def test_first_harmonic_amplitude_zonal_wind():

    tmp_amp = fft_wind_retrieval.first_harmonic_amplitude(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert np.isclose(fft_wind_retrieval.zonal_wind(tmp_amp).values[1, 1], 0)


def test_first_harmonic_amplitude_projected_wind():

    tmp_amp = fft_wind_retrieval.first_harmonic_amplitude(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert np.isclose(
        fft_wind_retrieval.wind_projected_to_azimuth(tmp_amp, 0).values[1, 1],
        -20,
    )


def test_first_harmonic_amplitude_horizontal_wind_speed_var():

    tmp_ds = fft_wind_retrieval.get_wind_properties(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert "horizontal_wind_speed" in tmp_ds


def test_first_harmonic_amplitude_horizontal_wind_dir_var():

    tmp_ds = fft_wind_retrieval.get_wind_properties(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert "horizontal_wind_direction" in tmp_ds


def test_first_harmonic_amplitude_meridional_wind_dir_var():

    tmp_ds = fft_wind_retrieval.get_wind_properties(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert "meridional_wind" in tmp_ds


def test_first_harmonic_amplitude_zonal_wind_dir_var():

    tmp_ds = fft_wind_retrieval.get_wind_properties(
        get_radial_velocities_4_test().radial_wind_speed
    )
    assert "zonal_wind" in tmp_ds



import numpy as np
import pandas as pd
import xarray as xr

from lidarwind.io import open_sweep


def wc_azimuth_elevation_correction(
    ds: xr.Dataset, azimuth_resolution: int = 1, elevation_resolution: int = 1
):

    """Azimuth and elevation correction

    This function corrects the azimuth ambiguity issue by
    replacing the 360 azimuth value with 0. It also rounds
    the elevation and azimuth according to a specified
    resolution

    Parameters
    ----------
    ds : xr.Dataset
        A dataset containing the WindCube's observations

    azimuth_resolution : int
        Azimuth resolution in degrees

    elevation_resolution : int
        Elevation resolution in degrees

    Returns
    -------
    xr.Dataset
        The same dataset, but the azimuth is corrected for the ambiguity

    """

    assert "azimuth" in ds
    assert ds["azimuth"].dims == ("time",)

    ds["azimuth"] = ds["azimuth"].round(azimuth_resolution)
    # Avoid ambiguity on 360 degrees
    ds["azimuth"] = ds["azimuth"].where(ds.azimuth != 360, 0)

    ds["elevation"] = ds["elevation"].round(elevation_resolution)

    return ds


def wc_fixed_files_restruc_dataset(ds: xr.Dataset):

    """Restructuring fixed type files

    This function restructures the WindCube fixed
    file type; it is needed to allow the merging
    of several fixed-type files. The function starts
    by swapping the range dimension with the gate_index.
    Then, the time dimension is added to the variables
    that do not have it. After that, elevation, azimuth
    and range are defined as new coordinates.

    Parameters
    ----------
    ds : xr.Dataset
        A dataset from the WindCube's fixed file type

    Returns
    -------
    xr.Dataset
        The same dataset, but restructured for merging.

    """

    if not isinstance(ds, xr.Dataset):
        raise TypeError

    assert "time" in ds
    assert "range" in ds
    assert "azimuth" in ds
    assert "elevation" in ds
    assert "gate_index" in ds

    ds["gate_index"] = ds["gate_index"].astype("i")
    ds = ds.swap_dims({"range": "gate_index"}).reset_coords()

    tmp_no_time = ds[
        [v for v in ds.variables if "time" not in ds[v].dims]
    ].expand_dims("time")
    tmp_time = ds[[v for v in ds.variables if "time" in ds[v].dims]]

    ds = xr.merge([tmp_no_time, tmp_time])
    ds = ds.set_coords(["elevation", "azimuth", "range"])

    return ds


def wc_fixed_merge_files(file_names: list):

    """Merging fixed type files

    This function merges multiple fixed files into a single dataset.

    Parameters
    ----------
    file_names : list
        A list of fixed files to be merged

    Returns
    -------
    xr.Dataset
        A dataset containing data from all files specified
        in the file_names list

    """

    zenith_list = []
    slanted_list = []

    if bool(file_names) is False:
        raise FileNotFoundError

    for file in file_names:

        tmp_ds = wc_fixed_files_restruc_dataset(open_sweep(file))

        if tmp_ds["elevation"] == 90:
            zenith_list.extend([tmp_ds])

        if tmp_ds["elevation"] != 90:
            slanted_list.extend([tmp_ds])

        tmp_ds.close()

    ds_zenith = xr.merge(zenith_list)
    ds_slanted = xr.merge(slanted_list)

    ds = xr.merge([ds_zenith, ds_slanted])

    return ds


def wc_slanted_radial_velocity_4_fft(ds: xr.Dataset):

    """Extraction of slanted radial velocities

    This function extracts the non-zenith pointing
    data from the dataset. Only the non-zenith pointing
    data is used to retrieve the horizontal wind speed and
    direction using the FFT method. Additionally, it sets
    azimuth as a new dimension and reindexes the data by
    proximity in time. The function identifies the duration
    of a complete scan cycle, and for a given time step,
    the time index is used as a central reference and
    assigned as the time index of all azimuthal observations
    within the duration of one cycle.


    Parameters
    ----------
    ds : xr.Dataset
        A dataset of fixed-type files merged,
        corrected for azimuth and elevation ambiguity
        and from a single elevation.


    Rerturns
    --------
    xr.Dataset
        A dataset of reindexed slanted observations with
        an extra dimension: azimuth

    """

    if len(np.unique(ds.elevation)) > 1:
        raise TypeError("This dataset contains multiple elevations")

    if 90 in np.unique(ds.elevation):
        raise ValueError(
            "90 degrees elevation: not valid for retrieving horizontal wind"
        )

    if (
        len(ds["azimuth"].where(ds["azimuth"] == ds["azimuth"][0], drop=True))
        < 2
    ):
        raise ValueError(
            "Not enough data to estimate the one scan cylce duration"
        )

    # initializing storage ds
    radial_velocities = xr.Dataset()

    # identify the mean duration of a complete scan cycle
    half_cycle = (
        ds.where(ds.azimuth == ds.azimuth[0], drop=True)
        .time.diff(dim="time")
        .mean()
        .values
    )
    half_cycle = pd.to_timedelta(half_cycle).seconds / 2

    # unique azimuths
    azimuth = np.unique(ds.azimuth)

    for azm in azimuth:

        # selecting the reference azimuthal slice
        azimuth_left_over = azimuth[azimuth != azm]
        tmp_reference_slice = ds["radial_wind_speed"].where(
            ds["azimuth"] == azm, drop=True
        )
        tmp_reference_slice = (
            tmp_reference_slice.drop(["azimuth"])
            .assign_coords({"azimuth": azm})
            .expand_dims(["azimuth"])
        )
        radial_velocities = xr.merge([radial_velocities, tmp_reference_slice])

        for azm_left in azimuth_left_over:

            tmp_azimuth_slice = ds["radial_wind_speed"].where(
                ds["azimuth"] == azm_left, drop=True
            )
            interp_azimuth_slice = tmp_azimuth_slice.reindex(
                time=tmp_reference_slice.time,
                method="nearest",
                tolerance=f"{half_cycle}s",
            )
            tmp_slice = (
                interp_azimuth_slice.drop(["azimuth"])
                .assign_coords({"azimuth": azm_left})
                .expand_dims(["azimuth"])
            )
            radial_velocities = xr.merge([radial_velocities, tmp_slice])

    radial_velocities["azimuth"].attrs = ds["azimuth"].attrs

    return radial_velocities



import numpy as np
import pytest
import xarray as xr

import lidarwind
from lidarwind import preprocessing

files = lidarwind.sample_data("wc_6beam")
files = sorted(files)


def get_sintetic_6_beam_data():

    elv = np.array([75, 75, 90, 75, 75, 75, 75])
    data_elv = xr.DataArray(
        elv, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    azm = np.array([0, 72, 0, 144, 216, 288, 0])
    data_azm = xr.DataArray(
        azm, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    gate = np.array([104, 104, 100, 104, 104, 104, 104])
    data_range = xr.DataArray(
        gate, dims=("time"), coords={"time": np.arange(len(elv))}
    )

    data = xr.DataArray(
        np.array([1, 1, 1, 1, 1, 1, 1])[:, np.newaxis],
        dims=("time", "gate_index"),
        coords={
            "time": np.arange(len(elv)),
            "gate_index": [1],
        },
    )

    sintetic_data = xr.Dataset(
        {
            "range": data_range,
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr": data,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data,
        }
    )

    sintetic_data = sintetic_data.set_coords(["azimuth", "elevation", "range"])

    return sintetic_data


def test_wc_fixed_files_restruc_dataset_ds_type():

    with pytest.raises(TypeError):
        preprocessing.wc_fixed_files_restruc_dataset(ds=np.array([0, 1]))


def test_wc_fixed_files_restruc_dataset(file_name=files[0]):

    ds = lidarwind.open_sweep(file_name)
    ds = preprocessing.wc_fixed_files_restruc_dataset(ds)

    assert "time" in ds.dims
    assert "gate_index" in ds.dims

    assert "azimuth" in ds.coords
    assert "elevation" in ds.coords
    assert "gate_index" in ds.coords

    assert len(ds.variables) > 0

    ds.close()


def test_wc_fixed_merge_files_empty_file_names():

    with pytest.raises(FileNotFoundError):
        preprocessing.wc_fixed_merge_files(file_names=[])


# @pytest.fixture
def test_wc_fixed_merge_files_ds_structure(file_names=files[0:6]):

    ds = preprocessing.wc_fixed_merge_files(file_names)

    assert len(ds.time) > 1
    assert len(ds.gate_index) > 1

    ds.close()


# @pytest.mark.skip("skip 1")
def test_wc_slanted_radial_velocity_4_fft_90_deg_elevation():

    ds = get_sintetic_6_beam_data()
    ds = preprocessing.wc_azimuth_elevation_correction(ds)
    ds = ds.where(ds["elevation"] == 90, drop=True)

    assert ds["elevation"].size > 0

    with pytest.raises(ValueError):
        preprocessing.wc_slanted_radial_velocity_4_fft(ds)

    ds.close()


# @pytest.mark.skip("skip 1")
def test_wc_slanted_radial_velocity_4_fft_multiple_elevation():

    ds = get_sintetic_6_beam_data()
    ds = preprocessing.wc_azimuth_elevation_correction(ds)

    with pytest.raises(TypeError):
        preprocessing.wc_slanted_radial_velocity_4_fft(ds)

    ds.close()


# @pytest.mark.skip("skip 1")
def test_wc_slanted_radial_velocity_4_fft_few_data():

    ds = get_sintetic_6_beam_data()
    ds = preprocessing.wc_azimuth_elevation_correction(ds)
    unique_elevation = np.unique(ds.elevation)[np.unique(ds.elevation) != 90]

    with pytest.raises(ValueError):
        preprocessing.wc_slanted_radial_velocity_4_fft(
            ds.where(ds.elevation == unique_elevation, drop=True)
        )

    ds.close()



import numpy as np
import xarray as xr

from ..preprocessing.wind_cube import wc_slanted_radial_velocity_4_fft
from ..wind_retrieval.fft_wind_retrieval import get_wind_properties


def get_horizontal_wind(ds: xr.Dataset) -> xr.Dataset:
    """Apply fft retrieval

    This function applies the fft wind retrieval method
    to the preprocessed dataset

    Parameters
    ----------
    ds : xr.Dataset
        A dataset of preprocessed observations

    Returns
    -------
        The input dataset, but including the horizontal
        wind profiles.

    """

    unique_elevation = np.unique(ds.elevation)[np.unique(ds.elevation) != 90]
    ds_slanted = ds.where(ds.elevation == unique_elevation, drop=True)

    radial_velocities = wc_slanted_radial_velocity_4_fft(ds_slanted)
    horizontal_wind = get_wind_properties(radial_velocities.radial_wind_speed)

    ds = ds.merge(horizontal_wind)

    return ds


# Post processing
def wc_extract_wind(ds: xr.Dataset, method="full") -> xr.Dataset:
    """Wind profiles extraction

    This function extracts the wind information from
    the processed dataset.

    Parameters
    ----------
    ds : xr.Dataset
        A dataset processed and containing the retrieved
        horizontal wind.

    method : string
        Method for extracting the wind dataset.
        If "full", the data is merged, preserving all information.
        If "compact", the data is merged, but only the horizontal
        wind profiles that are almost simultaneous to the
        vertical observations are merged.

    Returns
    -------
        The wind dataset

    """

    vertical_velocity = ds.radial_wind_speed.where(
        ds["elevation"] == 90, drop=True
    )
    vertical_velocity.name = "vertical_wind_speed"
    vertical_velocity = (
        vertical_velocity.drop(["gate_index", "range"])
        .assign_coords(
            {"gate_index": vertical_velocity.range.isel(time=0).values}
        )
        .rename({"gate_index": "range"})
    )

    variables = [
        "horizontal_wind_speed",
        "horizontal_wind_direction",
        "meridional_wind",
        "zonal_wind",
    ]

    horizontal_wind = ds[variables]

    horizontal_wind = horizontal_wind.where(
        horizontal_wind["elevation"] != 90, drop=True
    )
    horizontal_wind = horizontal_wind.where(
        np.isfinite(horizontal_wind.range), drop=True
    )
    new_range = (
        np.sin(np.deg2rad(horizontal_wind.elevation)) * horizontal_wind.range
    )
    horizontal_wind = (
        horizontal_wind.drop(["gate_index", "range"])
        .assign_coords({"gate_index": new_range.isel(time=0).values})
        .rename({"gate_index": "range"})
    )

    horizontal_wind = horizontal_wind.interp(
        {"range": vertical_velocity.range}, method="nearest"
    )

    if method == "full":
        wind = horizontal_wind.merge(vertical_velocity)

    if method == "compact":

        wind = (
            horizontal_wind.interp(
                {"time": vertical_velocity["time"]}, method="nearest"
            )
            .drop(["elevation"])
            .merge(vertical_velocity)
        )

    wind = wind.drop(["azimuth", "azimuth_length", "freq_azimuth"])

    return wind



import pytest

import lidarwind as lst

from .data import data_filenames  # , get_sample_data


@pytest.fixture
# @pytest.mark.skip("deactivating to isolate sef fault error")
def test_DataOperations(data_filenames):

    tmp_object = lst.DataOperations(data_filenames)

    return tmp_object.merged_data


# @pytest.mark.skip("deactivating to isolate sef fault error")
def test_GetRestructuredData(test_DataOperations):

    lst.GetRestructuredData(test_DataOperations)



from ..wind_retrieval.fft_wind_retrieval import get_wind_properties


def get_horizontal_wind(ds):

    wind_ds = get_wind_properties(ds.MeanVel)
    wind_ds = wind_ds.merge(
        ds[
            [
                "start_scan",
                "end_scan",
                "zdr_max",
                "nan_percentual",
                "chirp_start",
                "chirp_end",
                "chirp_azimuth_bias",
                "azm_seq",
            ]
        ]
    )
    wind_ds = wind_ds.expand_dims(["mean_time"])

    return wind_ds



import numpy as np
import pandas as pd
import xarray as xr

from lidarwind import postprocessing, preprocessing


def sintetic_data(step=90, elevation=75) -> xr.Dataset:
    """Sintect data

    Function to create synthetic data for testing the package

    Parameters
    ----------
    step : int
        separation in degree between the data points

    elevation : int
        elevation of the observation

    Returns
    -------
        A synthetic dataset

    """

    azimuths = np.arange(0, 360, step)

    wind = np.cos(np.deg2rad(azimuths + 0)) * 20 * np.cos(np.deg2rad(75))

    elv = np.append(np.ones_like(azimuths) * elevation, [90, 75])

    time = pd.to_datetime("11/11/2017") + pd.to_timedelta(
        np.arange(len(elv)), unit="seconds"
    )

    data_elv = xr.DataArray(elv, dims=("time"), coords={"time": time})

    azm = np.append(azimuths, [0, 0])
    data_azm = xr.DataArray(azm, dims=("time"), coords={"time": time})

    data = xr.DataArray(
        (np.ones((2, 7)) * np.append(wind, [0, wind[0]])).T,
        dims=("time", "gate_index"),
        coords={"time": time, "gate_index": [1, 2]},
    )

    ranges = xr.DataArray(
        np.ones((2, 7)).T
        * np.array([100, 150])
        / np.sin(np.deg2rad(elevation)),
        dims=("time", "gate_index"),
        coords={"time": time, "gate_index": [1, 2]},
    )

    data_status = xr.DataArray(
        np.ones_like(data.values),
        dims=("time", "gate_index"),
        coords={"time": time, "gate_index": [1, 2]},
    )

    test_ds = xr.Dataset(
        {
            "elevation": data_elv,
            "azimuth": data_azm,
            "cnr": data,
            "range": ranges,
            "radial_wind_speed": data,
            "radial_wind_speed_status": data_status,
            "relative_beta": data,
        }
    )

    test_ds = test_ds.set_coords(
        {
            "elevation": test_ds.elevation,
            "azimuth": test_ds.azimuth,
            "range": ranges,
        }
    )

    test_ds["range"].values[test_ds.elevation == 90] = np.array([100, 150])

    return test_ds


def ds_for_test():

    ds = sintetic_data(step=72, elevation=75)
    ds = preprocessing.wc_azimuth_elevation_correction(ds)
    ds = ds.where(ds.radial_wind_speed_status == 1)

    return ds


def test_post_wind_cube_get_horizontal_wind_dim_size():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert len(tmp_ds.dims) > 0


def test_post_wind_cube_get_horizontal_wind_time():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "time" in tmp_ds.dims


def test_post_wind_cube_get_horizontal_wind_gate_index():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "gate_index" in tmp_ds.dims


def test_post_wind_cube_get_horizontal_wind_coord_range():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "range" in tmp_ds.coords


def test_post_wind_cube_get_horizontal_wind_coord_elevation():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "elevation" in tmp_ds.coords


def test_post_wind_cube_get_horizontal_wind_coord_azimuth():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "azimuth" in tmp_ds.coords


def test_post_wind_cube_get_horizontal_wind_speed():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "horizontal_wind_speed" in tmp_ds.variables


def test_post_wind_cube_get_horizontal_wind_direction():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "horizontal_wind_direction" in tmp_ds.variables


def test_post_wind_cube_get_horizontal_wind_zonal():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "zonal_wind" in tmp_ds.variables


def test_post_wind_cube_get_horizontal_wind_meridional():

    tmp_ds = postprocessing.get_horizontal_wind(ds_for_test())
    assert "meridional_wind" in tmp_ds.variables



import numpy as np
import xarray as xr
import xrft


def first_harmonic_amplitude(
    radial_velocity: xr.DataArray, dim="azimuth"
) -> xr.DataArray:
    """First harmonic amplitude

    This function calculates the complex amplitudes
    along the azimuth coordinate and returns
    the amplitude of the first harmonic.

    Parameters
    ----------
    radial_velocity : xr.DataArray
        A data array of slanted Doppler velocities

    dim : string
        Name of the azimuthal dimension

    Returns
    -------
        A data array of the amplitudes of the first harmonic

    """

    complex_amplitudes = xrft.fft(
        radial_velocity, dim=dim, true_amplitude=False
    )
    complex_amplitudes = complex_amplitudes.assign_coords(
        {f"{dim}_length": len(radial_velocity[dim])}
    )

    # determination of the first harmonic position
    # harmonic_index = int(radial_velocity[dim].size / 2) + 1
    first_harmonic_index = (
        complex_amplitudes[f"freq_{dim}"]
        .where(complex_amplitudes[f"freq_{dim}"] > 0)
        .argmin()
    )

    return complex_amplitudes.isel({f"freq_{dim}": first_harmonic_index})


def harmonic_phase(amplitude: xr.DataArray) -> xr.DataArray:
    """First harmonic phase

    It calculates the phase of a given complex harmonic amplitude.

    Parameters
    ----------
    amplitude : xr.DataArray
        A data array of complex amplitudes from a given harmonic

    Returns
    -------
        harmonic phase

    """

    phase = np.rad2deg(np.arctan2(amplitude.imag, amplitude.real))

    return phase


def wind_direction(amplitude: xr.DataArray) -> xr.DataArray:
    """wind direction

    It calculates the wind direction of a given complex harmonic amplitude.

    Parameters
    ----------
    amplitude : xr.DataArray
        A data array of complex amplitudes from a given harmonic

    Returns
    -------
        The wind direction

    """

    wind_direction = -harmonic_phase(amplitude) + 180

    return wind_direction


def wind_speed(
    amplitude: xr.DataArray, coord="elevation", azimuth_name="azimuth"
) -> xr.DataArray:
    """Wind speed

    It calculates the wind speed using the first harmonic.

    Parameters
    ----------
    amplitude : xr.DataArray
        A data array of complex amplitudes from the first harmonic

    coord : string
        Name of the elevation coordinate


    Returns
    -------
        The wind speed

    """

    radial_wind_speed = (
        2 * np.abs(amplitude) / amplitude[f"{azimuth_name}_length"]
    )
    horizontal_wind_speed = radial_wind_speed / np.cos(
        np.deg2rad(amplitude[coord])
    )

    return horizontal_wind_speed


def wind_projected_to_azimuth(
    amplitude: xr.DataArray,
    azimuth: float,
    coord="elevation",
    azimuth_name="azimuth",
) -> xr.DataArray:
    """Wind speed from an azimuth

    It calculates the wind speed from a given azimuth.

    Parameters
    ----------
    amplitude : xr.DataArray
        A data array of complex amplitudes from the first harmonic

    azimuth : float
        Azimuth value

    coord : string
        Name of the elevation coordinate

    Returns
    -------
        The projection of the horizontal

    """

    radial_wind_speed = (
        2 * np.abs(amplitude) / amplitude[f"{azimuth_name}_length"]
    )
    projected_radial_wind = radial_wind_speed * np.cos(
        np.deg2rad(azimuth) + np.deg2rad(-harmonic_phase(amplitude) + 180)
    )
    projected_horizontal_wind = projected_radial_wind / np.cos(
        np.deg2rad(amplitude[coord])
    )

    return projected_horizontal_wind


def zonal_wind(
    amplitude: xr.DataArray, elevation_name="elevation", azimuth_name="azimuth"
) -> xr.DataArray:
    """zonal wind

    It calculates the zonal wind component.

    Parameters
    ----------
    amplitude : xr.DataArray
        A data array of complex amplitudes from the first harmonic

    Returns
    -------
        The zonal wind component

    """

    zonal_wind = (
        wind_projected_to_azimuth(
            amplitude, 90, coord=elevation_name, azimuth_name=azimuth_name
        )
        * -1
    )

    return zonal_wind


def meridional_wind(
    amplitude: xr.DataArray, elevation_name="elevation", azimuth_name="azimuth"
) -> xr.DataArray:
    """meridional wind

    It calculates the meridional wind component.

    Parameters
    ----------
    amplitude : xr.DataArray
        A data array of complex amplitudes from the first harmonic

    Returns
    -------
        The meridional wind component

    """

    meridional_wind = (
        wind_projected_to_azimuth(
            amplitude, 0, coord=elevation_name, azimuth_name=azimuth_name
        )
        * -1
    )

    return meridional_wind


def get_wind_properties(
    radial_velocity: xr.DataArray,
    elevation_name="elevation",
    azimuth_name="azimuth",
) -> xr.Dataset:
    """Wind dataset

    It retrieves the wind properties from the slanted observations.

    Parameters
    ----------
    radial_velocities : xr.DataArray
        A data array of the slanted Doppler velocities observations.

    Returns
    -------
        A dataset containing the wind speed, direction, meridional
        and zonal components

    """

    amplitude = first_harmonic_amplitude(radial_velocity, dim=azimuth_name)

    wind_properties = xr.Dataset()
    wind_properties["horizontal_wind_direction"] = wind_direction(amplitude)
    wind_properties["horizontal_wind_speed"] = wind_speed(
        amplitude, coord=elevation_name, azimuth_name=azimuth_name
    )
    wind_properties["meridional_wind"] = meridional_wind(
        amplitude, elevation_name=elevation_name, azimuth_name=azimuth_name
    )
    wind_properties["zonal_wind"] = zonal_wind(
        amplitude, elevation_name=elevation_name, azimuth_name=azimuth_name
    )

    return wind_properties
